#!/usr/bin/perl -w

# Program bsparser.pl

# (c) Gary C. Kessler, 2011-2013 [gck@garykessler.net]

# Program to parse a FAT or NTFS Boot Sector
# Ref: File System Forensic Analysis, Brian Carrier,
# http://www.ntfs.com/ntfs-partition-boot-sector.htm,
# and other documentation

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

#use Encode qw/encode_utf8 decode_utf8/;
use utf8;
binmode STDOUT, ":utf8";

# Initialize global variables

$version = "3.1";
$build_date = "13 March 2013";

# $source is the name of the file to read. It is expected to be a
# 512-byte file. Initially, we don't know what type of partion we have.

$source = "";
$BStype = "";
$suppress = FALSE;

# boot_type is an array with the boot sector type

%boot_type =
  (
  "f" => "FAT12/16",
  "f32" => "FAT32",
  "n" => "NTFS"
  );

# media_type is an array with the media type labels

%media_type =
  (
  F0 => "Removable disk",
  F8 => "Fixed disk"
  );

#
# Begin program activity!!
#

print "\nBoot Sector Parser V$version - Gary C. Kessler ($build_date)\n\n";

# Parse command for file name and request missing information.

if (parsecommandline ())
  { exit 1; };

print "Source file = $source; Partition type = $boot_type{$BStype}\n\n";

# Open input ($source) file and stop program if error

open (INFILE, "<:raw", $source)
  or die "Can't open input file \"$source\": $!";
binmode INFILE;

# Start by reading the entire file, which should be 512 bytes

$i = 0;
while (not (eof INFILE) && $i < 512)
  {
  read (INFILE,$byte,1);
  $boot [$i] = unpack ("C", $byte);
  $i++;
  }

if ($i != 512)
  { die "File is smaller than one sector ($i bytes)\n"; }

case:
  {
  ($BStype eq "f" || $BStype eq "f32") && do
    {
    boot_sector_FAT_common ();
    };
  $BStype eq "n" && do
    {
    boot_sector_NTFS ();
    };
  }

# Finish up!

close INFILE;


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# boot_sector_FAT_common
# ********************************************************

# Process bytes 0-35, the common bytes for FAT12, FAT16,
# and FAT32 boot sectors

sub boot_sector_FAT_common
{
# Bytes 0-2 contain the instruction to jump to the boot code

$bpos = 0; $len = 3;
printf ("%03d-%03d  Jump instruction: ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Bytes 3-10 contain the OEM name in ASCII

$bpos = 3; $len = 8;
printf ("%03d-%03d  OEM: ", $bpos,$bpos+$len-1);
for ($i = $bpos; $i < $bpos+$len; $i++)
  { print chr($boot[$i]); }
print "\n";

# Bytes 11-12 contain the number of bytes/sector (little endian)

$bpos = 11; $len = 2;
printf ("%03d-%03d  Sector size: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " bytes]\n";

# Byte 13 contains the number of sectors/cluster

$bpos = 13;
printf ("%03d      Cluster size: 0x%02X [%d sectors]\n", $bpos, $boot[$bpos], $boot[$bpos]);

# Bytes 14-15 contain the number of reserved sectors

$bpos = 14; $len = 2;
printf ("%03d-%03d  Size of reserved space: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
printf (" [%d sector(s)]\n", $temp);

# Byte 16 contains the number of FATs

$bpos = 16;
printf ("%03d      Number of FATs: 0x%02X [%d]\n", $bpos, $boot[$bpos], $boot[$bpos]);

# Bytes 17-18 contain the number of files in Root Directory (FAT12/16 only);
# 0 for FAT32

$bpos = 17; $len = 2;
printf ("%03d-%03d  Root Directory entries: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]";

if ($temp != 0)
    { print " ** FAT12/FAT16 **\n"; }
  else
    { print " ** FAT32 **\n"; }

# Bytes 19-20 contain the number of sectors in the file system, only if the
# value fits in 2 bytes (usually systems <=32MB). If 2 bytes is insufficient,
# this value is 0 and the 4-byte value in bytes 32-35 is used instead.

$bpos = 19; $len = 2;
printf ("%03d-%03d  File system size (<32 MB): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify ($temp), " sectors] ";

if ($temp == 0)
    { print " ** See bytes 32-35 **\n"; }
  else
    { print "\n"; }

# Byte 21 is the media type

$bpos = 21;
$temp = sprintf ("%02X", $boot[$bpos]);
printf ("%03d      Media type: 0x%02X ", $bpos, $boot[$bpos]);
if (defined $media_type{$temp})
    { print "[$media_type{$temp}]\n"; }
  else
    { print "(Not valid)\n"; }

# Bytes 22-23 contain the number of sectors in each FAT (FAT12/16 only);
# 0 for FAT32

$bpos = 22; $len = 2;
printf ("%03d-%03d  FAT size: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " sectors] ";

if ($temp != 0)
    { print " ** FAT12/FAT16 **\n"; }
  else
    { print " ** FAT32 **\n"; }

# Bytes 24-25 contain the number of sectors per track

$bpos = 24; $len = 2;
printf ("%03d-%03d  Track size: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " sectors]\n";

# Bytes 26-27 contain the number of heads

$bpos = 26; $len = 2;
printf ("%03d-%03d  Number of heads: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 28-31 contain the number of sectors before the start partition

$bpos = 28; $len = 4;
printf ("%03d-%03d  Number of sectors before start: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 32-35 contain the number of sectors in the file system if the 2 byte
# value above is insufficient (usually systems >32MB).

$bpos = 32; $len = 4;
printf ("%03d-%03d  File system size (>32 MB): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " sectors] ";

if ($temp == 0)
    { print " ** See bytes 19-20 **\n"; }
  else
    { print "\n"; }

# End of common space! Now jump to FAT12/FAT16- and FAT32-specific routines

print "\n";

case:
  {
  $BStype eq "f" && do
    {
    boot_sector_FAT12_16 ();
    };
  $BStype eq "f32" && do
    {
    boot_sector_FAT32 ();
    };
  }

return;
}

# ********************************************************
# boot_sector_FAT12_16
# ********************************************************

# Process bytes 36-511 of a FAT12/FAT16 boot sector

sub boot_sector_FAT12_16
{
# Byte 36 contains the BIOS INT 13H drive number (hex 8x)

$bpos = 36;
printf ("%03d      INT 13H drive number: 0x%02X\n", $bpos, $boot[$bpos]);

# Byte 37 is not used

$bpos = 37;
printf ("%03d      (Unused): 0x%02X\n", $bpos, $boot[$bpos]);

# Byte 38 contains the extended boot signature (0x29)

$bpos = 38;
printf ("%03d      Extended boot signature: 0x%02X ", $bpos, $boot[$bpos]);

if ($boot[$bpos] == 0x29)
    { print "[valid]\n"; }
  else
    { print "[invalid]\n"; }

# Bytes 39-42 contain the volume serial number

$bpos = 39; $len = 4;
printf ("%03d-%03d  Volume serial number: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
printf (" [%d]\n", $temp);

# Bytes 43-53 contain the volume label

$bpos = 43; $len = 11;
printf ("%03d-%03d  Volume label: ", $bpos,$bpos+$len-1);
for ($i = $bpos; $i < $bpos+$len; $i++)
  { print chr($boot[$i]); }
print "\n";

# Bytes 54-61 contain the file system type

$bpos = 54; $len = 8;
printf ("%03d-%03d  File system type: ", $bpos,$bpos+$len-1);
for ($i = $bpos; $i < $bpos+$len; $i++)
  { print chr($boot[$i]); }
print "\n";

# Bytes 62-509 are boot code (print 28 rows of 16 entries) unless suppressed

$bpos = 62; $len = 448; 
printf ("\n%03d-%03d  Boot code", $bpos,$bpos+$len-1);

if ($suppress eq TRUE)
    { 
    print " (Remove the -s switch to display the boot code)\n";
    }
  else
    {
    print " (Use the -s switch to suppress this display)\n";

    for ($i=0; $i<=27; $i++)
      {
      set_output ($bpos,16,@boot);
      $bpos += 16;
      }
    }

print "\n";

# Finally, process the final two-byte signature (should be 0xAA-55)

$bpos = 510; $len = 2;
printf ("%03d-%03d  Signature: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp == 0xaa55)
    { print " [valid]\n\n"; }
  else
    { print " [invalid]\n\n"; }

return;
}

# ********************************************************
# boot_sector_FAT32
# ********************************************************

# Process bytes 36-511 of a FAT32 boot sector

sub boot_sector_FAT32
{
# Bytes 36-39 contain the size of one FAT

$bpos = 36; $len = 4;
printf ("%03d-%03d  Size of one FAT: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " sectors]\n";

# Bytes 40-41 describe how the multiple FATs are used

$bpos = 40; $len = 2;
printf ("%03d-%03d  FAT-usage descriptor: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print "\n         ";

# If bit 7 = 1, only one FAT is active and bits 0-3 indicate the
# active FAT number. Otherwise, all FATs are active.

if ($temp >= 128)
    { printf ("FAT table %d is active (bit 7 = 1)\n", $temp%16); }
  else
    { print "All FATs are active (bit 7 = 0)\n"; }

# Bytes 42-43 describe the major and minor version number

$bpos = 42; $len = 2;
printf ("%03d-%03d  Major.minor version number: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
printf (" [v%d.%d]\n", $boot [$bpos+1], $boot [$bpos]);

# Bytes 44-47 contain the cluster where the Root Directory can be found

$bpos = 44; $len = 4;
printf ("%03d-%03d  Root Directory start cluster: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 48-49 contain the sector where FSINFO can be found

$bpos = 48; $len = 2;
printf ("%03d-%03d  FSINFO sector address: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 50-51 contain the sector where the backup boot sector can be found

$bpos = 50; $len = 2;
printf ("%03d-%03d  Backup boot sector address: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]";
if ($temp == 6)
    { print " (Default)\n"; }
  else
    { print " (Non-default)\n"; }

# Bytes 52-63 are reserved

$bpos = 52; $len = 12;
printf ("%03d-%03d  (Reserved): ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Byte 64 contains the BIOS INT 13H drive number (hex 8x)

$bpos = 64;
printf ("%03d      INT 13H drive number: 0x%02X\n", $bpos, $boot[$bpos]);

# Byte 65 is not used

$bpos = 65;
printf ("%03d      (Unused): 0x%02X\n", $bpos, $boot[$bpos]);

# Byte 66 contains the extended boot signature (0x29)

$bpos = 66;
printf ("%03d      Extended boot signature: 0x%02X ", $bpos, $boot[$bpos]);

if ($boot[$bpos] == 0x29)
    { print "[valid]\n"; }
  else
    { print "[invalid]\n"; }

# Bytes 67-70 contain the volume serial number

$bpos = 67; $len = 4;
printf ("%03d-%03d  Volume serial number: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
printf (" [%d]\n", $temp);

# Bytes 71-81 contain the volume label

$bpos = 71; $len = 11;
printf ("%03d-%03d  Volume label: ", $bpos,$bpos+$len-1);
for ($i = $bpos; $i < $bpos+$len; $i++)
  { print chr($boot[$i]); }
print "\n";

# Bytes 82-89 contain the file system type

$bpos = 82; $len = 8;
printf ("%03d-%03d  File system type: ", $bpos,$bpos+$len-1);
for ($i = $bpos; $i < $bpos+$len; $i++)
  { print chr($boot[$i]); }
print "\n";

# Bytes 90-509 are unused (print 26 rows of 16 entries plus one row of 4
# entries) unless suppressed

$bpos = 90; $len = 420; 
printf ("\n%03d-%03d  (Unused)", $bpos,$bpos+$len-1);

if ($suppress eq TRUE)
    { 
    print " (Remove the -s switch to display these values)\n";
    }
  else
    {
    print " (Use the -s switch to suppress this display)\n";

    for ($i=0; $i<=25; $i++)
      {
      set_output ($bpos,16,@boot);
      $bpos += 16;
      }
    set_output ($bpos,4,@boot);
    }

print "\n";

# Finally, process the final two-byte signature (should be 0xAA-55)

$bpos = 510; $len = 2;
printf ("%03d-%03d  Signature: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp == 0xaa55)
    { print " [valid]\n\n"; }
  else
    { print " [invalid]\n\n"; }

return;
}

# ********************************************************
# boot_sector_NTFS
# ********************************************************

# Parse a NTFS partition's boot sector

sub boot_sector_NTFS
{
# Bytes 0-2 contain the instruction to jump to the boot code

$bpos = 0; $len = 3;
printf ("%03d-%03d  Jump instruction: ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Bytes 3-10 contain the OEM name in ASCII

$bpos = 3; $len = 8;
printf ("%03d-%03d  OEM: ", $bpos,$bpos+$len-1);
for ($i = $bpos; $i < $bpos+$len; $i++)
  { print chr($boot[$i]); }
print "\n";

# Bytes 11-12 contain the number of bytes/sector (little endian)

$bpos = 11; $len = 2;
printf ("%03d-%03d  Sector size: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " bytes]\n";

# Byte 13 contains the number of sectors/cluster

$bpos = 13;
printf ("%03d      Cluster size: 0x%02X [%d sectors]\n", $bpos, $boot[$bpos], $boot[$bpos]);

# Bytes 14-15 are reserved (and should be 0)

$bpos = 14; $len = 2;
printf ("%03d-%03d  (Reserved): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp != 0)
    { print " ** Invalid, per Microsoft **\n"; }
  else
    { print "\n"; }

# Bytes 16-20 are unused (and should be 0)

$bpos = 16; $len = 5;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp != 0)
    { print " ** Invalid, per Microsoft **\n"; }
  else
    { print "\n"; }

# Byte 21 is the media type

$bpos = 21;
$temp = sprintf ("%02X", $boot[$bpos]);
printf ("%03d      Media type: 0x%02X ", $bpos, $boot[$bpos]);
if (defined $media_type{$temp})
    { print "[$media_type{$temp}]\n"; }
  else
    { print "(Not valid)\n"; }

# Bytes 22-23 are unused (and should be 0)

$bpos = 22; $len = 2;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp != 0)
    { print " ** Invalid, per Microsoft **\n"; }
  else
    { print "\n"; }

# Carrier reports that bytes 24-31 are not used by NTFS. Other
# sources show what is here...

# Bytes 24-25 contain the number of sectors per track

$bpos = 24; $len = 2;
printf ("%03d-%03d  Track size: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " sectors]\n";

# Bytes 26-27 contain the number of heads

$bpos = 26; $len = 2;
printf ("%03d-%03d  Number of heads: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 28-31 contain the number of sectors before the start partition

$bpos = 28; $len = 4;
printf ("%03d-%03d  Number of sectors before start: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 32-35 are unused (and should be 0)

$bpos = 32; $len = 4;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp != 0)
    { print " ** Invalid, per Microsoft **\n"; }
  else
    { print "\n"; }

# Bytes 36-39 are unused, but always set to 0x80-00-80-00

$bpos = 36; $len = 4;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);

if ($temp != 0x800080)
    { print " ** Invalid, per Microsoft **\n"; }
  else
    { print "\n"; }

# Bytes 40-47 contain the number of sectors in the file system

$bpos = 40; $len = 8;
printf ("%03d-%03d  File system size: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), " sectors]\n";

# Bytes 48-55 contain the $MFT's starting logical cluster address

$bpos = 48; $len = 8;
printf ("%03d-%03d  \$MFT starting logical cluster: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Bytes 56-63 contain the $MFT Mirror's starting logical cluster address

$bpos = 56; $len = 8;
printf ("%03d-%03d  \$MFT Mirror starting logical cluster: ", $bpos,$bpos+$len-1);
$temp = little_endian ($bpos,$len,@boot);
list_hex_digits ($bpos, $len, @boot);
print " [", commify($temp), "]\n";

# Byte 64 contains the size of an MFT entry, in clusters

$bpos = 64;
printf ("%03d      Size of MFT entry: 0x%02X [%d clusters]\n", $bpos, $boot[$bpos], $boot[$bpos]);

# Bytes 65-67 are unused

$bpos = 65; $len = 3;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Byte 68 contains the size of an INDEX record (directory entry), in clusters

$bpos = 68;
printf ("%03d      Size of INDEX (directory) entry: 0x%02X [%d cluster(s)]\n", $bpos, $boot[$bpos], $boot[$bpos]);

# Bytes 69-71 are unused

$bpos = 69; $len = 3;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Bytes 72-79 are the volume serial number

$bpos = 72; $len = 8;
printf ("%03d-%03d  Volume serial number: ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Bytes 80-83 are unused

$bpos = 80; $len = 4;
printf ("%03d-%03d  (Unused): ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);
print "\n";

# Bytes 84-509 contain boot code (print 26 rows of 16 entries plus 1 row of
# 10 entries) unless suppressed

print "\nBytes 84-509 are the boot code";

if ($suppress eq TRUE)
    { 
    print "\n";
    }
  else
    {
    print " (Use the -s switch to suppress this display):\n";

    for ($i=0; $i<26; $i++)
      {
      $bpos = 84 + $i * 16;
      set_output ($bpos,16,@boot);
      }
    $bpos = 84 + 26 * 16;
    set_output ($bpos,10,@boot);
    }

print "\n";

# Finally, process the final two-byte signature (should be 0xAA-55)

$bpos = 510; $len = 2;
$temp = little_endian ($bpos,$len,@boot);
printf ("%03d-%03d  Signature: ", $bpos,$bpos+$len-1);
list_hex_digits ($bpos, $len, @boot);

if ($temp == 0xaa55)
    { print " [valid]\n\n"; }
  else
    { print " [invalid]\n\n"; }

return;
}


# ********************************************************
# commify
# ********************************************************

# Routine from PerlFAQ5 -- add commas to large integers
# http://perldoc.perl.org/perlfaq5.html
  #How-can-I-output-my-numbers-with-commas-added?

sub commify

{
local $_  = shift;
1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
return $_;
}

# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: bsparser [-i input_file] [-f | -f32 | -n] {-s}
               bsparser {[-h] | [-v] | [-tf] | [-tn]}

 where: -i   is the input file name
        -f   indicates a FAT12 or FAT16 boot sector file
        -f32 indicates a FAT32 boot sector file
        -n indicates an NTFS boot sector file
        -s will suppress display of the boot code block
        -h prints this help file
        -v displays the program version number
        -tf prints the FAT boot sector template file
        -tn prints the NTFS bot sector template file

If no -i switch is provided, the program will ask for the input
file name.

If neither -f, -f32, nor -n switches are provided, the program
will prompt for the boot sector type.

The input file should contain the contents of a Boot Sector. The
file should be a binary file that is a single sector (512 bytes)
in length. The best way to obtain the B/S is to export the file
from your forensics software. (The Boot Sector is sometimes called
a Volume Boot Record.)

EOT
return;
}


# ********************************************************
# list_hex_digits
# ********************************************************

# Given an array name, starting position, and number of digits, print
# out a string of hex digits in the format 0xAA-BB-CC...

sub list_hex_digits ($$@)
{
my ($start,$n,@array) = @_;
my $i;

for ($i = $start; $i < $start+$n; $i++)
  {
  if ($i == $start)
      { print "0x"; }
    else
      { print "-"; }
  printf ("%02X", $array[$i]);
  }

return;
}

# ********************************************************
# little_endian
# ********************************************************

# Given an array name, starting position, and number of digits,
# calculate the decimal value using little endian (LSB first)

sub little_endian ($$@)
{
my ($start,$n,@array) = @_;
my ($i, $sum);

$sum = $array[$start];
for ($i=1; $i<$n; $i++)
  { $sum += $array[$start+$i] * (2**($i*8)); }

return $sum;
}


# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present. Query
# user for any missing information

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)
  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      ($ARGV[$i] eq "-f" || $ARGV[$i] eq "-n") && do
         {
         $BStype = substr ($ARGV[$i],1,1);
         last PARMS;
         };
      $ARGV[$i] eq "-f32" && do
         {
         $BStype = substr ($ARGV[$i],1,3);
         last PARMS;
         };
      ($ARGV[$i] eq "-tf" || $ARGV[$i] eq "-tn") && do
         {
         if ($ARGV[$i] eq "-tf")
             { $layout = "layout_BS_FAT.txt"; }
            else
             { $layout = "layout_BS_NTFS.txt"; }
         open (INFILE, "<", $layout)
          or die "Can't open input file \"$layout\": $!";
         while (not (eof INFILE))
           {
           $line = <INFILE>;
           print $line;
           }
         close INFILE;
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-s" && do
         {
         $suppress = TRUE;
         last PARMS;
         };
      $ARGV[$i] eq "-v" && do
         {
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: bsparser [-i input_file] [-f | -n] {-s}\n";
         print "       bsparser {[-h] | [-v] | [-tf] | [-tn]}\n\n";
         $state = 1;
         return $state;
         };
      };
    };
  };

# Prompt for file names, if not supplied

if ($source eq "")
  {
  print "Enter the input file name: ";
  chomp ($source = <STDIN>);
  print "\n";
  };

# Prompt for boot sector type, if not supplied

if ($BStype eq "")
  {
  while ($BStype ne "f" && $BStype ne "n" && $BStype ne "f32")
    {
    print "Enter the boot sector type (f = FAT12/16, f32 = FAT32, or n = NTFS): ";
    chomp ($BStype = <STDIN>);
    print "\n";
    }
  };

return $state;
}


# ********************************************************
# set_output
# ********************************************************

# Print output of the data bytes to show both hex and ASCII
# values.

sub set_output ($$@)
{
my ($bpos,$n,@array) = @_;
my ($j, $k, $linemax);

# Format for a maximum of 16 items per line

$linemax = 16;

printf ("%03d: ", $bpos);
for ($j=0; $j<$n; $j++)
  {
  printf (" %02X", $array[$bpos+$j]);
  }
if ($n < $linemax)
  {
  for ($j=1; $j<=$linemax-$n; $j++)
    { print "   "; };
  }

print "   ";
for ($j=0; $j<$n; $j++)
  {
  $k = $array[$bpos+$j];

#   Print a dot (.) for unprintable characters

  if ($k <= 0x1f || ($k >= 0x7f && $k <= 0x9f))
      { print "."; }
    else
      { print chr($k); }
  }
print "\n";

return;
}

